from __future__ import annotations
from abc import abstractmethod
from typing_extensions import *
from typing import List, Dict, overload
import copy
import uuid
from uuid import UUID
import numpy as np

from manim import *


class DynamicMobjectGraph():

    def __init__(self):
        self.root_mobjects: Set[MobjectIdentity] = []
        self.subscriptions: Dict[UUID, Callable[[DynamicMobjectGraph], None]] = {}
        self.id = None
    
    @property
    def mobjects(self) -> List[MobjectIdentity]:
        mobjects: Set[MobjectIdentity] = set()

        for root_mobject in self.root_mobjects:
            connected_mobjects = self.connected_from_root(root_mobject)
            for mobject in connected_mobjects:
                mobjects.add(mobject)
        
        return list(mobjects)

    @property
    def dynamic_mobjects(self) -> List[DynamicMobject]:
        return [ mobject.current_dynamic_mobject for mobject in self.mobjects ]
    
    def contains(self, id: UUID) -> bool:
        for mobject in self.dynamic_mobjects:
            if mobject.id == id:
                return True
            
        return False
    
    def find_dynamic_mobject(self, id: UUID) -> DynamicMobject | None:
        for mobject in self.dynamic_mobjects:
            if mobject.id == id:
                return mobject
            
        return None
    
    def get_dynamic_mobject(self, id: UUID) -> DynamicMobject:
        mobject = self.find_dynamic_mobject(id)

        if mobject is None:
            raise Exception(f"No mobject with id {id} found in graph")
        
        return mobject
    
    @staticmethod
    def connected_from_root(root_mobject: MobjectIdentity) -> Set[MobjectIdentity]:
        mobjects: Set[MobjectIdentity] = set()

        def recursive_extract(mobject: MobjectIdentity):
            mobjects.add(mobject)
            for child in mobject.children:
                recursive_extract(child)

        recursive_extract(root_mobject)
        return mobjects


    def subscribe(self, function: Callable[[DynamicMobjectGraph], None], id: UUID | None = None) -> UUID:
        if id is None:
            id = uuid.uuid4()
        self.subscriptions[id] = function
        return id
    
    def unsubscribe(self, id: UUID):
        del self.subscriptions[id]

    def copy(self) -> DynamicMobjectGraph:
        subscriptions = self.subscriptions
        self.subscriptions = {}
        graph = copy.deepcopy(self)
        self.subscriptions = subscriptions
        return graph
    

    def begin_invalidation(self):
        for function in list(self.subscriptions.values()).copy():
            function(self)

    def begin_state_invalidation(self):

        for function in list(self.subscriptions.values()).copy():
            function(self)

        for mobject in self.mobjects:
            try:
                center = mobject.current_dynamic_mobject.get_center()
                mobject.mobject_center = center
            except:
                pass
    
    
    def set_root_mobjects(self, root_mobjects: Set[MobjectIdentity]):

        for next_root in root_mobjects:
            if next_root.parent is not None:
                raise Exception(
                    f"Attempted set_root_mobjects` call on mobject {next_root}, that is a child of mobject {next_root.parent}."
                    "A root mobject cannot have another mobject as its parent."
                )

        for curr_root in self.root_mobjects.copy():
            if curr_root not in root_mobjects:

                self.root_mobjects.remove(curr_root)
                
                graph = DynamicMobjectGraph()
                graph.root_mobjects = { curr_root }
                
                curr_root.graph = graph
        
        for next_root in root_mobjects:
            if next_root not in self.root_mobjects:
                
                next_root.graph.root_mobjects.remove(next_root)

                self.root_mobjects.add(next_root)

                next_root.graph = self

    def set_root(self, mobject: MobjectIdentity):
        self.set_root_mobjects({ mobject })

    def add_root(self, mobject: MobjectIdentity):
        self.set_root_mobjects({ *self.root_mobjects, mobject })

    def remove_root(self, mobject: MobjectIdentity):
        next_root_mobjects = self.root_mobjects.copy()
        next_root_mobjects.remove(mobject)
        self.set_root_mobjects(next_root_mobjects)
        

    def connect_parent_child(self, parent: MobjectIdentity, child: MobjectIdentity):  

        graph1 = parent.graph
        graph2 = child.graph

        """
        if graph1 is graph2:
            graph1.root_mobjects.remove(child) ###
            child.graph = None
        else: """
        if graph1 is not graph2:
            root_connected_mobjects1 = graph1.mobjects
            root_connected_mobjects2 = graph2.mobjects

            m = {}
            for mobject in root_connected_mobjects1:
                m[mobject.id] = mobject

            for mobject in root_connected_mobjects2:
                if mobject.id in m:
                    raise Exception(
                        "Mobject graph cannot contain duplicate ids."
                        "Use mobject.clone() instead of mobject.copy() for duplication."
                    )
            
            graph2_root_mobjects = graph2.root_mobjects.copy()
            graph2.root_mobjects = set()
            graph1.root_mobjects = graph1.root_mobjects.union(graph2_root_mobjects)

        graph1.root_mobjects.remove(child) ###
        child.graph = None

        parent.children.add(child)
        child.parent = parent

    def disconnect_parent_child(self, parent: MobjectIdentity, child: MobjectIdentity):

        parent.children.remove(child)
        child.parent = None

        graph = DynamicMobjectGraph()
        graph.root_mobjects = { child }
        child.graph = graph


class MobjectIdentity(VMobject):

    def __init__(self, mobject: DynamicMobject):

        super().__init__()
        self.id = uuid.uuid4()
        self.source_ids: List[UUID] = []
        self.target_ids: List[UUID] = [] 


        self.parent: MobjectIdentity | None = None
        self.children: Set[MobjectIdentity] = set()

        self.incorporated = False

        self.mobject_graph: DynamicMobjectGraph | None = None
        self.mobject_graph = DynamicMobjectGraph()
        self.mobject_graph.root_mobjects = { self }

        self.terminate_propogation_mobject: MobjectIdentity | None = None
        self.permit_propogation: bool = True
        self.change_parent_mobject: MobjectIdentity | None = None
        self.change_parent_mobject_replacement: DynamicMobject | None = None
    
        self.mobject_center = VMobject().get_center()
        
        self.current: DynamicMobject | None = None
        self.set_dynamic_mobject(mobject)

    def descendants(self) -> List[MobjectIdentity]:
        descendants: List[MobjectIdentity] = []

        def recursive_extract(mobject: MobjectIdentity):
            descendants.append(mobject)
            for child in mobject.children:
                recursive_extract(child)

        return descendants

    @property
    def current_dynamic_mobject(self) -> DynamicMobject:
        if self.current is not None:
            return self.current
        raise Exception()
    
    @current_dynamic_mobject.setter
    def current_dynamic_mobject(self, mobject: DynamicMobject):
        self.current = mobject
   
    @property
    def root_parent(self) -> MobjectIdentity:
        if self.parent is not None:
            return self.parent.root_parent
        else:
            return self
    
    def is_root(self) -> bool:
        return self.root_parent is self
    
    @property
    def graph(self) -> DynamicMobjectGraph:
        graph = self.root_parent.mobject_graph
        if graph is None:
            raise Exception()
        return graph

    @graph.setter
    def graph(self, graph: DynamicMobjectGraph):
        if self.root_parent is not self:
            raise Exception()
        self.mobject_graph = graph
    

    def set_dynamic_mobject(self, mobject: DynamicMobject):

        if mobject.mobject_identity is not None and mobject.mobject_identity is not self:
            raise Exception()
        
        if self.current is not None:
            self.graph.begin_state_invalidation()
        
        if self.current is not None and self.current is not mobject:
            #self.graph.begin_state_invalidation()
            self.current.submobjects = [ self ]

        #self.graph.begin_invalidation()
        #self.graph.begin_state_invalidation()

        self.current = mobject
        self.current.mobject_identity = self
        self.submobjects = [ self.current ]

        self.begin_entrance_invalidation()

    def begin_entrance_invalidation(self):
        self.invalidate(terminate_propogation_mobject=self.terminate_propogation_mobject)
        self.terminate_propogation_mobject = None

    def complete_child_registration(self):
        self.set_children(self.next_children)

    def invalidate(
            self, 
            terminate_propogation_mobject: MobjectIdentity | None = None
        ):
        self.next_children: List[MobjectIdentity] = []
        self.current_dynamic_mobject.execute_compose()

        if self.parent is not None and self.parent is terminate_propogation_mobject:
            # terminate propogation
            pass
        else:
            self.invalidate_parent()

    def invalidate_parent(self):
        if self.parent is not None:
            self.parent.invalidate()
        else:
            pass
    
    def conditional_clone(self, mobject: DynamicMobject) -> DynamicMobject:

        if mobject.identity in self.next_children:
            return mobject.clone()
        
        if mobject.identity is self.change_parent_mobject:
            replacement = self.change_parent_mobject_replacement
            self.change_parent_mobject = None
            self.change_parent_mobject_replacement = None
            return replacement
        
        return mobject

    def register_child(self, mobject: Mobject):

        if isinstance(mobject, DynamicMobject):
            mobject = self.conditional_clone(mobject)
            self.next_children.append(mobject.identity)
            mobject = mobject.identity.current_dynamic_mobject
        
        return mobject
    
    def set_children(self, next_children: List[MobjectIdentity]):
        
        for identity in self.children.copy():
            if identity not in next_children:
                MobjectIdentity.remove_parent_connection(parent=self, child=identity)

        for identity in next_children:
            if identity not in self.children.copy():
                MobjectIdentity.add_parent_connection(parent=self, child=identity)
            
    @staticmethod
    def add_parent_connection(parent: MobjectIdentity, child: MobjectIdentity):
        
        if child.parent is not None:
            child.graph.begin_invalidation()
            child.parent.change_parent_mobject = child
            child.parent.change_parent_mobject_replacement = child.current_dynamic_mobject.clone()
            child.parent.terminate_propogation_mobject = parent
            child.parent.graph.begin_state_invalidation()

            child_parent = child.parent
            child.parent.invalidate()
            child_parent.terminate_propogation_mobject = None
        
        parent.graph.connect_parent_child(parent, child)
    
    @staticmethod
    def remove_parent_connection(parent: MobjectIdentity, child: MobjectIdentity):
        parent.graph.disconnect_parent_child(parent, child)

    def __str__(self) -> str:
        return f"{self.current_dynamic_mobject}-{str(self.id).split('-')[0]}"


MobjectType = TypeVar("MobjectType")
DynamicMobjectType = TypeVar("DynamicMobjectType")

class DynamicMobjectSubgraph(VMobject):

    @property
    def dynamic_mobjects(self) -> List[DynamicMobject]:
        return [ mobject.current_dynamic_mobject for mobject in self.mobjects ]

    def __init__(self, *dynamic_mobjects: DynamicMobject, graph: DynamicMobjectGraph | None = None):

        if not dynamic_mobjects and graph is None:
            raise Exception()
        
        if graph is None:
            graph = dynamic_mobjects[0].graph
        
        for mobject in dynamic_mobjects:
            if mobject.graph is not graph:
                raise Exception()
        
        self.mobjects: Set[MobjectIdentity] = [ mobject.identity for mobject in dynamic_mobjects ]
        self.graph = graph
        super().__init__()
        self.submobjects = [ mobject.direct_submobject_tree() for mobject in self.dynamic_mobjects ]

    def contains(self, id: UUID):
        for mobject in self.mobjects:
            if mobject.id == id:
                return True
        
        return False

    def find_dynamic_mobject(self, id: UUID) -> DynamicMobject:
        if not self.contains(id):
            return None
        else:
            return self.graph.find_dynamic_mobject(id)
        
    def get_dynamic_mobject(self, id: UUID) -> DynamicMobject:
        if not self.contains(id):
            raise Exception(f"No mobject with id {id} found in subgraph")
        else:
            return self.graph.find_dynamic_mobject(id)

    @staticmethod
    def from_adapt(mobject: Mobject) -> DynamicMobjectSubgraph:
        
        if not isinstance(mobject, DynamicMobjectSubgraph):
            if not isinstance(mobject, DynamicMobject):
                raise Exception()

            mobject = DynamicMobjectSubgraph.from_dynamic_mobject(mobject)

        return mobject
    
    @classmethod
    def from_dynamic_mobject(cls, mobject: DynamicMobject) -> DynamicMobjectSubgraph:
        return cls(*mobject.get_dynamic_family())
    
    def __sub__(self, mobject: Mobject) -> DynamicMobjectSubgraph:
        subgraph = DynamicMobjectSubgraph.from_adapt(mobject)
        dynamic_mobjects = set(self.dynamic_mobjects).difference(set(subgraph.dynamic_mobjects))
        return DynamicMobjectSubgraph(*dynamic_mobjects)

    def __add__(self, mobject: Mobject) -> DynamicMobjectSubgraph:
        subgraph = DynamicMobjectSubgraph.from_adapt(mobject)
        dynamic_mobjects = set(self.dynamic_mobjects).union(set(subgraph.dynamic_mobjects))
        return DynamicMobjectSubgraph(*dynamic_mobjects)



class DynamicMobject(VMobject):

    def __init__(
        self,
        id: UUID | None = None, 
        **kwargs
    ):
        self.shift_flag = False
        self.scale_flag = False
        self.scale_factor = 1
        self.position_factor = VMobject().get_center()

        self._save_x: float | None = None
        self._save_y: float | None = None
        self.arrange_function = None

        self.in_composite_edit = False

        super().__init__(**kwargs)
        self.mobject_identity: MobjectIdentity | None = None
        
        MobjectIdentity(self)

        if id is not None:
            self.identity.id = id
    
    def execute_compose(self):
        
        self.in_compose = True
        self.shift_during_compose_flag = False
        self.scale_during_compose_flag = False

        mobject_encoding = self.compose()
        self.identity.complete_child_registration()
        self.in_compose = False

        if mobject_encoding:
            if isinstance(mobject_encoding, list):
                self.submobjects = mobject_encoding
            elif isinstance(mobject_encoding, Mobject):
                self.submobjects = [ mobject_encoding ]
            else:
                raise TypeError(
                    f"{self.__class__.__name__}.compose() returned unsupported type {type(mobject_encoding)}"
                )
        
        if not self.scale_during_compose_flag:
            self.restore_scale()

        """
        if self.parent is None:
            if not self.shift_during_compose_flag:
                self.move_to(self.identity.mobject_center) """
        
        if self.arrange_function is not None:
            self.arrange_function(VGroup(*self.submobjects))
        
        if not self.shift_during_compose_flag:
            self.move_to(self.identity.mobject_center) 
            
    @abstractmethod
    def compose(self) -> None | Mobject | List[Mobject]:
        pass
    
    def is_current_dynamic_mobject_guard(self):
        if self.identity.current_dynamic_mobject is not self:
            raise Exception(
                f"Cannot modify former handle for mobject-${self.identity.id}"
                f"which has currently represented by {self.identity.current_dynamic_mobject}"
            )

    def invalidate(self) -> Self:
        self.is_current_dynamic_mobject_guard()  
        self.identity.set_dynamic_mobject(self)
        return self

    def restore_scale(self):
        super().scale(self.scale_factor)

    def accept_submobjects(self, *mobject: Mobject):
        self.submobjects = [ *mobject ]

    def pop(self):
        self.parent.remove(self)
        return self

    @abstractmethod
    def remove(self, mobject: DynamicMobject):
        raise NotImplementedError()

    def __add__(self, mobject: Mobject) -> DynamicMobjectSubgraph:
        return DynamicMobjectSubgraph.from_dynamic_mobject(self) + mobject
    
    def __sub__(self, mobject: Mobject) -> DynamicMobjectSubgraph:
        return DynamicMobjectSubgraph.from_dynamic_mobject(self) - mobject
    
    def subgraph(self) -> DynamicMobjectSubgraph:
        return DynamicMobjectSubgraph.from_dynamic_mobject(self)

    def direct_submobject_tree(self) -> Mobject:
        
        def recursive_extract(mobject: Mobject, group: Mobject):

            for submobject in mobject.submobjects:
                if not isinstance(submobject, DynamicMobject):
                    if submobject.has_points():
                        group.add(submobject)
                    else:
                        subgroup = VGroup()
                        recursive_extract(submobject, subgroup)
                        group.add(subgroup)

        group = VGroup()
        recursive_extract(self, group)
        return group

    def get_point_dynamic_mobject(self):
        
        mobject_copy = self.copy()
        center = self.get_center()

        def convert_to_point_mobject(mobject: Mobject):

            def recursive_extract(mobject: Mobject, group):

                for submobject in mobject.submobjects:
                    if submobject.has_points():
                        point_mobject = submobject.get_point_mobject()
                        group.add(point_mobject)
                    else:
                        subgroup = VGroup()
                        recursive_extract(submobject, subgroup)
                        group.add(subgroup)

            group = VGroup()
            recursive_extract(mobject, group)
            return group

        def recursive_extract(mobject: DynamicMobject) -> DynamicMobject:

            direct = convert_to_point_mobject(mobject.direct_submobject_tree())

            mobject = mobject.become(
                DGroup(direct, *mobject.children)
            )

            for child in mobject.children:
                recursive_extract(child)

            return mobject

        mobject_copy = recursive_extract(mobject_copy)

        restore_submobjects = {}

        for mobject in set(mobject_copy.get_family()):
            restore_submobjects[mobject] = mobject.submobjects
            mobject.submobjects = []
            mobject.move_to(center)
        
        for mobject, submobjects in restore_submobjects.items():
            mobject.submobjects = submobjects

        return mobject_copy

    def __disconnect_graph_objects(self):

        mobject_graph = self.mobject_identity.mobject_graph

        self.mobject_identity.mobject_graph = None

        def restore_graph_objects():
            self.mobject_identity.mobject_graph = mobject_graph

        self.restore_graph_objects_function = restore_graph_objects

    def __restore_graph_objects(self):
        self.restore_graph_objects_function()

    
    def copy(self) -> DynamicMobject:
        
        dynamic_family = self.get_dynamic_family()

        for mobject in dynamic_family:
            mobject.__disconnect_graph_objects()

        parent = self.mobject_identity.parent
        self.mobject_identity.parent = None

        copy_mobject = super().copy()
        
        if not isinstance(copy_mobject, DynamicMobject):
            raise Exception()

        self.mobject_identity.parent = parent

        for mobject in dynamic_family:
            mobject.__restore_graph_objects()
        
        copy_graph = DynamicMobjectGraph()
        copy_graph.root_mobjects = { copy_mobject.identity }
        copy_mobject.identity.mobject_graph = copy_graph

        return copy_mobject
        
    def clone(self) -> DynamicMobject:
        copy_mobject = self.copy()

        for mobject in copy_mobject.get_dynamic_family():
            mobject.source_id = mobject.id
            mobject.id = uuid.uuid4()

        return copy_mobject
    
    def become(self, mobject: DynamicMobjectType) -> DynamicMobjectType:
        self.is_current_dynamic_mobject_guard()

        if not isinstance(mobject, DynamicMobject):
            raise TypeError(
                f"{type(self).__name__}.become() recieved a standard mobject."
                f"To use a standard mobject, for example, a Square(), try using `${type(self).__name__}.become(Dynamic(Square()))`"
            )

        mobject_centers = {}
        for vertex in mobject.get_dynamic_family():
            mobject_centers[vertex] = vertex.get_center()

        mobject1, mobject2 = mobject, mobject.clone()

        children1 = mobject1.children.copy()
        children2 = mobject2.children.copy()

        m = {}
        for child in children2:
            b = False
            for mobject in children1:
                if mobject.id == child.source_id:
                    m[child] = mobject
                    b = True
            if not b:
                raise Exception()

        for child in children1:
            mobject1.identity.change_parent_mobject = child.identity
            mobject1.identity.change_parent_mobject_replacement = child.clone()
            mobject1.identity.invalidate()

        for child in children2:
            mobject2.identity.change_parent_mobject = child.identity
            mobject2.identity.change_parent_mobject_replacement = m[child]
            mobject2.identity.invalidate()
        
        empty = DynamicMobject()
        empty.mobject_identity = None

        mobject2.identity.set_dynamic_mobject(empty)
        mobject2.mobject_identity = None
        mobject2.mobject_identity = self.mobject_identity

        for child in children1:
            child.identity.mobject_center = mobject_centers[child]

        self.mobject_identity.set_dynamic_mobject(mobject2)

        for child in children1:
            child.identity.mobject_center = mobject_centers[child]
        return mobject2

    def get_dynamic_family(self) -> List[DynamicMobject]:
        family: Set[DynamicMobject] = set()

        def recursive_extract(mobject: DynamicMobject):
            family.add(mobject)
            for child in mobject.children:
                recursive_extract(child)

        recursive_extract(self)
        return list(family)
        
 
    def shift(self, *vectors) -> Self:
        super().shift(*vectors)
        
        if self.in_compose:
            self.shift_during_compose_flag = True

        return self

    def scale(self, scale_factor: float, **kwargs) -> Self:
        self.scale_factor *= scale_factor
        super().scale(self.scale_factor, **kwargs)
        return self
    
    def register_child(self, mobject: DynamicMobject) -> DynamicMobject:
        return self.identity.register_child(mobject)

    def save_x(self):
        self._save_x = self.get_x()

    def save_y(self):
        self._save_y = self.get_y()

    def save_center(self):
        self.save_x()
        self.save_y()

    def restore_x(self, propogate=True):
        factor = self._save_x - self.get_x()
        if propogate:
            self.root_parent.shift(np.array([ factor, 0, 0 ]))
        else:
            self.shift(np.array([ factor, 0, 0 ]))
        return self
    
    def restore_y(self, propogate=True) -> DynamicMobject:
        factor = self._save_y - self.get_y()
        if propogate:
            self.root_parent.shift(np.array([ 0, factor, 0 ]))
        else:
            self.shift(np.array([ 0, factor, 0 ]))
        return self
    
    def restore_center(self, propogate=True):
        self.restore_x(propogate)
        self.restore_y(propogate)

    @property
    def identity(self) -> MobjectIdentity:
        if self.mobject_identity is not None:
            return self.mobject_identity
        
        raise Exception()
    
    @property
    def parent(self) -> DynamicMobject | None:
        if self.identity.parent is not None:
            return self.identity.parent.current_dynamic_mobject
        
        return None
    
    @property
    def children(self) -> List[DynamicMobject]:
        return [ identity.current_dynamic_mobject for identity in self.mobject_identity.children ]

    @property
    def root_parent(self) -> DynamicMobject:
        return self.identity.root_parent.current_dynamic_mobject
    
    def is_root(self):
        return self.identity.is_root()

    @property
    def id(self) -> UUID:
        return self.identity.id
    
    @id.setter
    def id(self, id: UUID):
        self.identity.id = id

    @property
    def source_id(self) -> UUID | None:
        if not self.identity.source_ids:
            return None
        
        return self.identity.source_ids[-1]
    
    @source_id.setter
    def source_id(self, id: UUID):
        self.identity.source_ids.append(id)

    @property
    def target_id(self) -> UUID | None:
        if not self.identity.target_ids:
            return None
        
        return self.identity.target_ids[-1]
    
    @target_id.setter
    def target_id(self, id: UUID):
        self.identity.target_ids.append(id)

    @property
    def graph(self) -> DynamicMobjectGraph:
        return self.identity.graph
    
    def arrange(self, direction = RIGHT, buff = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER, center = True, **kwargs) -> Self:
        self.arrange_function = lambda mobject: mobject.arrange(direction, buff, center, **kwargs)
        self.invalidate()
        return self

    def clear_arrange_function(self):
        self.arrange_function = None
        self.invalidate()

    def begin_composite_edit(self):
        for mobject in self.get_dynamic_family():
            mobject.in_composite_edit = True

    def end_composite_edit(self):
        for mobject in self.get_dynamic_family():
            mobject.in_composite_edit = False

    
    def set_color(
        self, color: ParsableManimColor = YELLOW_C, family: bool = True
    ) -> Self:
        
        if not self.in_composite_edit:
            self.begin_composite_edit()
            self.invalidate()
            super().set_color(color=color, family=family)
            self.invalidate()
            self.end_composite_edit()
        else:
            super().set_color(color=color, family=family)

        return self

def extract_direct_dynamic_mobjects(mobject: Mobject):
    dynamic_mobjects: Set[DynamicMobject] = set()

    def recursive_extract(mobject):
        for submobject in mobject.submobjects:
            if isinstance(submobject, DynamicMobject):
                dynamic_mobjects.add(submobject)
            else:
                recursive_extract(submobject)

    if isinstance(mobject, DynamicMobject):
        return [ mobject ]
    else:
        recursive_extract(mobject)
        return list(dynamic_mobjects)
    

class DGroup(DynamicMobject):

    def __init__(
        self,
        *mobjects: VMobject,
        **kwargs
    ):
        self._mobjects = list(mobjects)
        super().__init__(**kwargs)

    def compose(self):
        self.dynamic_mobjects: Set[DynamicMobject] = set()

        def recursive_extract(mobject):
            for submobject in mobject.submobjects.copy():
                if isinstance(submobject, DynamicMobject):
                    self.dynamic_mobjects.add(submobject)
                else:
                    recursive_extract(submobject)
            
            mobject.submobjects = [ self.register_child(submobject) for submobject in mobject.submobjects ]
        
        group = VGroup(*self._mobjects)
        recursive_extract(group)
        self._mobjects = group.submobjects
        self.submobjects = group.submobjects
    
    def add(self, *mobjects: VMobject) -> Self:

        for m in mobjects:
            if not isinstance(m, Mobject):
                raise TypeError(f"All submobjects must be of type Mobject")
            if m is self:
                raise ValueError("Mobject cannot contain self")

        unique_mobjects = remove_list_redundancies(mobjects)
        if len(mobjects) != len(unique_mobjects):
            logger.warning(
                "Attempted adding some Mobject as a child more than once, "
                "this is not possible. Repetitions are ignored.",
            )

        self._mobjects = list_update(self._mobjects, unique_mobjects)
        self.invalidate()
        return self

    def remove(self, *mobjects: Mobject) -> Self:
        
        for mobject in mobjects:
            if mobject in self._mobjects:
                self._mobjects.remove(mobject)
        
        self.invalidate()
        return self
    
    @property
    def mobjects(self) -> List[Mobject]:
        return self._mobjects

    @mobjects.setter
    def mobjects(self, mobjects: List[Mobject]):
        self._mobjects = mobjects
        self.invalidate()


from __future__ import annotations
from abc import abstractmethod
from typing_extensions import *
from typing import List, Dict
import uuid
import math

import manim
from manim import *
from .dynamic_mobject import DynamicMobject
from .numpy_mobject_array import NumpyMobjectArray, map_2d

def pairwise(iterable):
    
    iterator = iter(iterable)
    try:
        prev = next(iterator)
    except StopIteration:
        return
    for item in iterator:
        yield prev, item
        prev = item


class MathEncodable(DynamicMobject):

    def __init__(
        self,
        *args,
        **kwargs
    ):
        self.in_compose = False
        self._tex_string = None
        super().__init__(*args, **kwargs)

    def execute_compose(self):
        
        math_encoding = self.compose_tex_string()
        self.identity.complete_child_registration()

        if isinstance(math_encoding, list):
            self.child_components = math_encoding
            self.tex_string = self.arg_separator.join([ child.tex_string for child in math_encoding ])
        else:
            self.tex_string = math_encoding

        if self.parent and isinstance(self.parent, MathEncodable):
            pass
        else:
            math_tex = self.render_tex_string(self.tex_string)
            self.accept_mobject_from_rendered_tex_string(math_tex)

        if not isinstance(self.parent, MathEncodable):
            self.restore_scale()

        #if self.parent is None:
        self.move_to(self.identity.mobject_center)

    @abstractmethod
    def compose_tex_string() -> str | List[MathEncodable]:
        pass

    @abstractmethod
    def get_tex_string(self) -> str:
        return self.tex_string

    """
    @abstractmethod
    def rendered_tex_string_submobject_count(self, mobject: VMobject) -> int:
        return len(SingleStringMathTex(self.get_tex_string()))
    """

    @abstractmethod
    def accept_mobject_from_rendered_tex_string(self, mobject: VMobject) -> int:
        pass
    
    def render_tex_string(self, tex_string: str) -> SingleStringMathTex:
        return SingleStringMathTex(tex_string)

    def __str__(self) -> str:
        if self.tex_string is None:
            return f"{self.__class__.__name__}(pre-init)"
        else:
            return f"{self.__class__.__name__}({self.tex_string})"


class MathComponent(MathEncodable):

    def adapt_input(self, term: Any) -> Any:

        if isinstance(term, MathEncodable):
            return term
        
        if isinstance(term, (str, int, float)):
            return MathString(str(term))
        
        if isinstance(term, List):
            return MathTex(*term) 
        
        """
        if isinstance(term, Mobject):
            return DefaultMobjectAdapter(term)
        
        return TypeError("term passed into MathComponent recieved type ", type(term))
        """
        return term

    def register_child(self, term: Any, ignore_type: bool = False) -> MathEncodable:
        
        if term is None:
            if self.permit_none_children:
                return term
            
            raise Exception()

        if not isinstance(term, MathEncodable) and not ignore_type:
            if isinstance(term, manim.MathTex):
                raise Exception("Use reactive_manim.MathTex instead of manim.MathTex")
            raise Exception("Recieved non-MathEncodble input of type ", type(term))
        
        term = super().register_child(term)
        return term

    def __init__(
        self,
        permit_none_children: bool = False,
        **kwargs,
    ):
        self.arg_separator = " "
        self.permit_none_children = permit_none_children
        super().__init__(**kwargs)
    
    @abstractmethod
    def compose_tex_string(self) -> str | List[MathEncodable]:
        pass

    def accept_mobject_from_rendered_tex_string(self, mobject: VMobject) -> int:
        submobject_count = self.distribute_rendered_tex_string_to_child_components(mobject)
        self.set_submobjects()
        return submobject_count

    def distribute_rendered_tex_string_to_child_components(self, mobject: VMobject) -> int:

        total_change_in_width = 0
        curr_index = 0

        for _, term in enumerate(self.child_components):
            
            #if isinstance(term, MathStringFragment):
            #    term_submobject_count = term.submobject_count
            #else:
            #    term_submobject_count = len(SingleStringMathTex(term.tex_string)) 

            """
            term_submobject_count = term.rendered_tex_string_submobject_count()

            new_index = curr_index + term_submobject_count + len("".join(self.arg_separator.split()))
            rendered_tex_string_portion = math_tex[curr_index:new_index]
            """

            term_submobject_count = term.accept_mobject_from_rendered_tex_string(mobject[curr_index:])
            new_index = curr_index + term_submobject_count + len("".join(self.arg_separator.split()))
            rendered_tex_string_portion = mobject[curr_index:new_index]

            if term.family_members_with_points():

                shift = term.get_left()[0] - rendered_tex_string_portion.get_left()[0]
                term.set_x(term.get_x() - shift)

                self.align_child(term, rendered_tex_string_portion)

                #term.align_to(rendered_tex_string_portion.get_left(), LEFT)
                #term.align_to(rendered_tex_string_portion.get_bottom(), DOWN)

                change_in_width = term.width - rendered_tex_string_portion.width # change_in_width=0 in standard cases
                terms_right = VGroup(*mobject[new_index:]) 
                terms_right.shift(RIGHT * change_in_width)
                total_change_in_width += change_in_width
            
            curr_index = new_index

        """
        center_displacement = total_change_in_width / 2
        
        for term in self.child_components:
            term.shift(LEFT * center_displacement)"""

        return curr_index

    def set_submobjects(self):
        
        submobjects = []
        for term in self.child_components:
            if term.has_no_points() and not term.submobjects:
                continue

            submobjects.append(term)
        self.submobjects = submobjects
    
    def align_child(self, term: VMobject, rendered_tex_string: VMobject):
        term.set_y(rendered_tex_string.get_y())
    

class MathTex(MathComponent):

    def __init__(
        self, 
        *terms: Any,
        **kwargs
    ):
        self._terms = self.adapt_terms(terms)
        super().__init__(**kwargs)

    def compose_tex_string(self):
        self._terms = [ self.register_child(term) for term in self._terms ]
        return self._terms
    
    def adapt_terms(self, terms):
        return [ self.adapt_input(term) for term in terms ]

    @property
    def terms(self):
        return self._terms

    @terms.setter
    def terms(self, terms: List[Any]):
        self._terms = self.adapt_terms(terms)
        self.invalidate()

    def insert(self, index: int, term):
        self._terms.insert(index, self.adapt_input(term))
        self.invalidate()

    def append(self, term):
        self._terms.append(self.adapt_input(term))
        self.invalidate()

    def remove(self, term):
        self._terms.remove(term)
        self.invalidate()

    def __len__(self) -> int:
        return len(self._terms)

    def __getitem__(self, index: int):
        return self._terms[index]

    def __setitem__(self, index: int, term):
        self._terms[index] = self.adapt_input(term)
        self.invalidate()

    def __iter__(self):
        return iter(self._terms)

class MathString(MathEncodable):

    def __init__(self, tex_string: str, **kwargs):
        self.tex_string = tex_string
        
        mobject = SingleStringMathTex(tex_string)
        self.submobject_group = mobject
        self.submobjects = mobject.submobjects
        super().__init__(**kwargs)

    def compose_tex_string(self) -> str:
        return self.tex_string

    def set_tex_string(self, tex_string: str) -> Self:
        self.tex_string = tex_string
        self.invalidate()
        return self

    def accept_mobject_from_rendered_tex_string(self, mobject: VMobject) -> int:

        submobject_count = len(SingleStringMathTex(self.tex_string))
        submobjects = mobject.submobjects[:submobject_count]
        self.submobject_group = VGroup(*submobjects).match_style(self.submobject_group)

        self.submobjects =  [ *submobjects ]
        return submobject_count


class MathStringFragment(MathEncodable):

    def __init__(self, tex_string: str, submobject_count: int = 0):
        self.tex_string = tex_string
        self.submobject_count = submobject_count
        self.first = True

        super().__init__()

    def execute_compose(self):

        if self.parent:
            super().execute_compose()

    def compose_tex_string(self):
        return self.tex_string
    
    def accept_mobject_from_rendered_tex_string(self, mobject: VMobject) -> int:

        if self.submobject_count == 0:
            self.submobjects = [ ]
        else:
            group = mobject[:self.submobject_count]

            if self.first == False:
                group.match_style(self)
            else:
                self.first = False
            
            self.submobjects = group.submobjects
        
        return self.submobject_count

class HSpaceTex(MathString):

    def __init__(self, buff):
        if isinstance(buff, (float, int)):
            super().__init__(f"\\hspace{{{buff}em}}")
        else:
            super().__init__(f"\\hspace{{{buff}}}")


class MathSequence(MathComponent):

    def __init__(
        self, 
        *sequence_terms,
        include_commas = True,
        extra_buff = 1,
    ):
        self.sequence_terms = [ self.adapt_input(term) for term in sequence_terms ]
        self.include_commas = include_commas
        self.extra_buff = extra_buff

        self.commas = []
        if self.include_commas:
            for _ in range(1, len(self.sequence_terms)):
                self.commas.append(MathString(","))

        super().__init__()

    def compose_tex_string(self):

        self.sequence_terms = [ self.register_child(term) for term in self.sequence_terms ]
        self.commas = [ self.register_child(comma) for comma in self.commas ]
        spacers = []

        components = []
        if self.include_commas:
            if self.sequence_terms:
                components.append(self.sequence_terms[0])
            for comma, term in zip(self.commas, self.sequence_terms[1:]):
                x = HSpaceTex(self.extra_buff)
                components += [ comma, x, term ]
                spacers += [x]
        else:
            if self.sequence_terms:
                components.append(self.sequence_terms[0])
            for term in self.sequence_terms[1:]:
                x = HSpaceTex(self.extra_buff)
                components += [ x, term ]
                spacers += [x]
        
        self.spacers = spacers
        return components

    def adapt_terms(self, terms):
        return [ self.adapt_input(term) for term in terms ]

    @property
    def terms(self):
        return self.sequence_terms

    @terms.setter
    def terms(self, terms: List[Any]):
        self.sequence_terms = self.adapt_terms(terms)
        self.invalidate()

    def remove(self, term):

        for index, _term in enumerate(self.sequence_terms):
            if _term is term:
                break

        self.sequence_terms.pop(index)
        
        if self.commas:
            if index == len(self.commas):
                index -= 1
            self.commas.pop(index)

        self.invalidate()

    def insert(self, index, term):
        term = self.adapt_input(term)

        if self.sequence_terms:
            self.commas.insert(index, MathString(","))

        self.sequence_terms.insert(index, term)
        self.invalidate()

    def append(self, term):
        self.insert(len(self), term)

    def __len__(self) -> int:
        return len(self.sequence_terms)

    def __getitem__(self, index: int):
        return self.sequence_terms[index]

    def __setitem__(self, index: int, term):
        self.sequence_terms[index] = self.adapt_input(term)
        self.invalidate()

    def __iter__(self):
        return iter(self.sequence_terms)


class MathList(MathSequence):

    def __init__(
        self,
        *sequence_terms,
        include_commas = True,
        extra_inner_buff = 0.5,
        extra_outer_buff = 0.35,
        bracket_l = "[",
        bracket_r = "]"
    ):
        self.bracket_l = MathString(bracket_l)
        self.bracket_r = MathString(bracket_r)
        self.extra_inner_buff = extra_inner_buff
        self.extra_outer_buff = extra_outer_buff
        
        super().__init__(*sequence_terms, include_commas=include_commas, extra_buff=extra_inner_buff)

    def compose_tex_string(self):

        self.sequence_terms = [ self.register_child(term) for term in self.sequence_terms ]
        self.commas = [ self.register_child(comma) for comma in self.commas ]
        spacers = []

        self.bracket_l = self.register_child(self.bracket_l)
        self.bracket_r = self.register_child(self.bracket_r)
        
        components = []
        components.append(self.bracket_l)
        components.append(HSpaceTex(self.extra_outer_buff))
        if self.include_commas:
            if self.sequence_terms:
                components.append(self.sequence_terms[0])
            for comma, term in zip(self.commas, self.sequence_terms[1:]):
                x = HSpaceTex(self.extra_inner_buff)
                components += [ comma, x, term ]
                spacers += [x]
        else:
            if self.sequence_terms:
                components.append(self.sequence_terms[0])
            for term in self.sequence_terms[1:]:
                x = HSpaceTex(self.extra_inner_buff)
                components += [ x, term ]
                spacers += [x]
        
        components.append(HSpaceTex(self.extra_outer_buff))
        components.append(self.bracket_r)
        self.spacers = spacers
        return components


class Term(MathComponent):

    def __init__(
            self, 
            term, 
            superscript = None, 
            subscript = None,
        ):
        
        self.base = self.adapt_input(term)
        self._subscript = self.adapt_input(subscript)
        self._superscript = self.adapt_input(superscript)

        self.bracket1 = MathString(r"{")
        self.bracket2 = MathString(r"}^")
        self.bracket3 = MathString(r"{")
        self.bracket4 = MathString(r"}_")
        self.bracket5 = MathString(r"{")
        self.bracket6 = MathString(r"}")

        super().__init__(permit_none_children=True)

    def compose_tex_string(self):

        self.base = self.register_child(self.base)
        self._subscript = self.register_child(self._subscript)
        self._superscript = self.register_child(self._superscript)

        if self._superscript is None and self._subscript is None:
            return [ self.base ]
        
        if self._superscript is None:
            return ([ 
                self.bracket1, 
                self.base, 
                self.bracket2, 
                self.bracket3, 
                self.bracket4, 
                self.bracket5, 
                self._subscript, 
                self.bracket6 
            ])

        if self._subscript is None:
            return ([ 
                self.bracket1, 
                self.base, 
                self.bracket2, 
                self.bracket3, 
                self._superscript, 
                self.bracket4, 
                self.bracket5, 
                self.bracket6 
            ])
        
        return ([ 
            self.bracket1, 
            self.base, 
            self.bracket2, 
            self.bracket3, 
            self._superscript, 
            self.bracket4, 
            self.bracket5, 
            self._subscript, 
            self.bracket6 
        ])

    def align_child(self, term, rendered_tex_string):

        if term is self.superscript:
            delta = term.get_bottom()[1] - rendered_tex_string.get_bottom()[1]
            term.set_y(term.get_y() - delta)

        if term is self.subscript:
            delta = term.get_top()[1] - rendered_tex_string.get_top()[1]
            term.set_y(term.get_y() - delta)

    @property
    def term(self):
        return self.base
    
    @term.setter
    def term(self, term: Any):
        self.base = self.adapt_input(term)
        self.invalidate()
    
    @property
    def subscript(self):
        return self._subscript
    
    @subscript.setter
    def subscript(self, subscript: Any):
        self._subscript = self.adapt_input(subscript)
        self.invalidate()
    
    @property
    def superscript(self):
        return self._superscript
    
    @superscript.setter
    def superscript(self, superscript: Any):
        self._superscript = self.adapt_input(superscript)
        self.invalidate()

    def remove(self, mobject):

        if mobject is self.term:
            self.term = MathString("")

        if mobject is self._superscript:
            self._superscript = None

        if mobject is self._subscript:
            self._subscript = None
        
        self.invalidate()


class Parentheses(MathComponent):

    def __init__(
        self,
        inner
    ):
        self._inner = self.adapt_input(inner)
        self.spacer = MathString("\mspace{-3mu}")
        self.bracket_l = BracketMathStringFragment(r"\left(")
        self.bracket_r = BracketMathStringFragment(r"\right)")
        super().__init__()

    def compose_tex_string(self):

        self._inner = self.register_child(self._inner)
        self.spacer = self.register_child(self.spacer)
        self.bracket_l = self.register_child(self.bracket_l)
        self.bracket_r = self.register_child(self.bracket_r)

        return ([
            self.spacer,
            self.bracket_l,
            self._inner,
            self.bracket_r
        ])
    
    @property
    def inner(self):
        return self._inner
    
    @inner.setter
    def inner(self, inner: Any):
        self._inner = self.adapt_input(inner)
        self.invalidate()


    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int):
        return self._inner[index]

    def __setitem__(self, index: int, term):
        self._inner[index] = self.adapt_input(term)
        self.invalidate()
        

class Function(MathComponent):

    def __init__(
        self,
        name,
        input
    ):
        self._name = self.adapt_input(name)
        self._input = self.adapt_input(input)
        self._parentheses = Parentheses(self._input)
        super().__init__()

    def compose_tex_string(self):
        self._name = self.register_child(self._name)
        self._parentheses = self.register_child(self._parentheses)
        self._input = self._parentheses._inner

        return ([
            self._name,
            self._parentheses
        ])
    
    @property
    def parentheses(self):
        return self._parentheses

    @property
    def input(self):
        return self._input
    
    @input.setter
    def input(self, input: Any):
        self._input = self.adapt_input(input)
        self.invalidate()

    @property
    def function_name(self):
        return self._name

    @function_name.setter
    def function_name(self, name):
        self._name = self.adapt_input(name)
        self.invalidate()


class Fraction(MathComponent):

    def __init__(
        self,
        numerator,
        denominator
    ):
        self._numerator = self.adapt_input(numerator)
        self._vinculum = MathStringFragment("", submobject_count=1)
        self._denominator = self.adapt_input(denominator)
        super().__init__()

    def compose_tex_string(self):
        self._numerator = self.register_child(self._numerator)
        self._vinculum = self.register_child(self._vinculum)
        self._denominator = self.register_child(self._denominator)

        self.child_components = [ self._numerator, self._vinculum, self._denominator ]
        return r"\frac{" + self._numerator.get_tex_string() + r"}{" + self._denominator.get_tex_string() + r"}"

    @property
    def numerator(self):
        return self._numerator

    @numerator.setter
    def numerator(self, numerator):
        self._numerator = self.adapt_input(numerator)
        self.invalidate()
    
    @property
    def vinculum(self):
        return self._vinculum
    
    @property
    def denominator(self):
        return self._denominator

    @denominator.setter
    def denominator(self, denominator):
        self._denominator = self.adapt_input(denominator)
        self.invalidate()


class BracketMathStringFragment(MathStringFragment):

    def __init__(
        self,
        tex_string: str = ""
    ):
        self.first_render = True
        super().__init__(tex_string)

    def bracket_length(self, mobject: VMobject):

        if len(mobject.submobjects) == 1:
            return 1
        
        bracket_submobjects = []
        
        for submobject1, submobject2 in pairwise(mobject.submobjects):
            bracket_submobjects.append(submobject1)

            bottom_points = []
            top_points = []

            for point in submobject1[0].points:
                if math.isclose(point[1], submobject1.get_bottom()[1]):
                    bottom_points.append(point)

            for point in submobject2[0].points:
                if math.isclose(point[1], submobject2.get_top()[1]):
                    top_points.append(point)

            min_bx = min(bottom_points, key=lambda point: point[0])[0]
            max_bx = max(bottom_points, key=lambda point: point[0])[0]
            min_tx = min(top_points, key=lambda point: point[0])[0]
            max_tx = max(top_points, key=lambda point: point[0])[0]

            def overlap(interval1, interval2):
                a1, a2 = interval1
                b1, b2 = interval2

                overlap_start = max(a1, b1)
                overlap_end = min(a2, b2)

                overlap_length = max(0, overlap_end - overlap_start)

                total_length = max(a2, b2) - min(a1, b1)

                if total_length == 0:
                    return 0.0 
                percentage_overlap = (overlap_length / total_length)

                return percentage_overlap

            ov = overlap([min_bx, max_bx], [min_tx, max_tx])
            if ov < 0.99:
                break

            dist_y = submobject2.get_top()[1] - submobject1.get_bottom()[1]
            dist_x = max_bx - min_bx 
            
            if dist_y < 0:
                if -dist_y/dist_x > 0.15:
                    break

            if submobject2 is mobject.submobjects[-1]:
                bracket_submobjects.append(submobject2)

        return len(bracket_submobjects)
    
    def accept_mobject_from_rendered_tex_string(self, mobject: VMobject) -> int:
        submobject_count = self.bracket_length(mobject)

        if submobject_count == 0:
            self.submobjects = []
        else:
            group = mobject[:submobject_count]

            if not self.first_render:
                group.match_style(self)
            else:
                self.first_render = False
        
            self.submobjects = [ *group ]

        return submobject_count


class MathCases(MathComponent):

    def __init__(
        self,
        *lines: MathEncodable
    ):
        self._lines = self.adapt_lines(lines)
        self.begin_cases_fragment = BracketMathStringFragment(r"\begin{cases}") #MathStringFragment(r"\begin{cases}", 1)
        self.end_cases_fragment = MathStringFragment(r"\end{cases}", 0)
        self.replacement_bracket = None
        super().__init__()

    def compose_tex_string(self):
        self._lines = [ self.register_child(line) for line in self._lines ]
        self.begin_cases_fragment = self.register_child(self.begin_cases_fragment)

        space = r"\left\{ \vphantom{\begin{cases}"
        for line in self._lines:
            space += r"\\"
        space += r"\end{cases}} \right."

        self.replacement_bracket = self.render_tex_string(space)

        processed_lines = []

        for index, line in enumerate(self._lines):
            processed_lines.append(line)
            if index < len(self.lines) - 1:
                fragment = MathStringFragment(r"\\", 0)
                processed_lines.append(fragment)

        return ([
            self.begin_cases_fragment,
            *processed_lines,
            self.end_cases_fragment
        ])
    
    def adapt_lines(self, lines):
        return [ self.adapt_input(line) for line in lines ]
    
    @property
    def bracket(self) -> MathEncodable:
        return self.begin_cases_fragment
    
    @property
    def lines(self) -> List[MathEncodable]:
        return self._lines
    
    @lines.setter
    def lines(self, lines: List[Any]):
        self._lines = self.adapt_lines(lines)
        self.invalidate()

    def insert(self, index: int, line):
        self._lines.insert(index, self.adapt_input(line))
        self.invalidate()

    def append(self, line):
        self._lines.append(self.adapt_input(line))
        self.invalidate()

    def remove(self, line):
        self._lines.remove(line)
        self.invalidate()
    
    def __len__(self) -> int:
        return len(self._lines)

    def __getitem__(self, index: int):
        return self._lines[index]

    def __setitem__(self, index: int, line):
        self._lines[index] = self.adapt_input(line)
        self.invalidate()

    def __iter__(self):
        return iter(self._lines)
    

class CaseLine(MathComponent):

    def __init__(
        self,
        output,
        condition,
    ):
        self._output = self.adapt_input(output)
        self._condition = self.adapt_input(condition)
        super().__init__()

    def compose_tex_string(self):
        self._output = self.register_child(self._output)
        self._condition = self.register_child(self._condition)

        return [ self._output, MathStringFragment(r"&", 0), self._condition ]
    
    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output):
        self._output = self.adapt_input(output)
        self.invalidate()

    @property
    def condition(self):
        return self._condition

    @condition.setter
    def condition(self, condition):
        self._condition = self.adapt_input(condition)
        self.invalidate()
    

class MathMatrix(MathComponent):

    def __init__(
        self,
        matrix: List[List[Any]]
    ):
        self.bracket_l = BracketMathStringFragment(r"\begin{bmatrix}")
        self.bracket_r = BracketMathStringFragment(r"\end{bmatrix}")

        self._matrix = NumpyMobjectArray.from_mobjects(
            map_2d(matrix, lambda elem: self.adapt_input(elem))
        )
        super().__init__()

    def compose_tex_string(self):

        if not self._matrix.is_2d():
            raise Exception()

        self.bracket_l = self.register_child(self.bracket_l)   
        self._matrix = NumpyMobjectArray.from_mobjects(
            map_2d(self._matrix.tolist(), lambda elem: self.register_child(elem) )
        )
        self.bracket_r = self.register_child(self.bracket_r)

        def mobject_row_encoding(mobject_row):
            encoding = []
            front_mobject, *next_mobjects = mobject_row

            if front_mobject:
                encoding += [ front_mobject ]

            for mobject in next_mobjects:
                encoding += [ MathStringFragment(r"&", 0), mobject ]

            return encoding
        
        encoding = []
        front_row, *next_rows = self._matrix.tolist()

        if front_row:
            encoding += [ *mobject_row_encoding(front_row) ]

        for row in next_rows:
            encoding += [ MathStringFragment(r"\\", 0), *mobject_row_encoding(row) ]

        return ([
            self.bracket_l,
            *encoding,
            self.bracket_r
        ])

    @property
    def matrix(self):
        return self._matrix.tolist()
    
    @matrix.setter
    def matrix(self, matrix: List[List[Any]]):
        self._matrix = NumpyMobjectArray.from_mobjects(
            map_2d(matrix, lambda elem: self.adapt_input(elem))
        )
        self.invalidate()

    def __len__(self) -> int:
        return len(self._matrix.tolist())

    def __getitem__(self, key):
        return self._matrix.tolist().__getitem__(key)

    def __iter__(self):
        return iter(self._matrix.tolist())


from __future__ import annotations
from abc import abstractmethod
from typing import Dict
from uuid import UUID

from manim import Scene
from .manim_src.composition import PreviousAnimationGroup
from .dynamic_mobject import *


class GraphProgressManager():

    def __init__(
        self,
        graph: DynamicMobjectGraph,
        scene: Scene
    ):
        self.scene = scene
        self.graph = graph
        self.graph.subscribe(lambda graph: self.end_transforms())
        self.is_transforming = False

        self.source_graph: DynamicMobjectGraph | None = None
        self.target_graph: DynamicMobjectGraph | None = None

        self.source_mobjects: Dict[UUID, MobjectIdentity] = {}
        self.target_mobjects: Dict[UUID, MobjectIdentity] = {}
        self.mobject_union: Dict[UUID, MobjectIdentity] = {}

        self.active_mobjects = []


    def begin_transforms(self):
        
        if self.is_transforming == True:
            return
        
        self.is_transforming = True
        
        for mobject in self.graph.root_mobjects:
            self.scene.scene_add(mobject.current_dynamic_mobject)
        
        self.target_graph = self.graph.copy() 
        self.target_mobjects = { mobject.id: mobject for mobject in self.graph.mobjects }
        
        for mobject in self.graph.mobjects:
            if mobject not in self.active_mobjects:
                self.active_mobjects.append(mobject)

        id_union = { id for id in self.source_mobjects.keys() }.union({ id for id in self.target_mobjects.keys() })
        self.mobject_union = {}

        for id in id_union:
            if id in self.target_mobjects:
                self.mobject_union[id] = self.target_mobjects[id]
            else:
                self.mobject_union[id] = self.source_mobjects[id]

        for mobject in self.mobject_union.values():
            mobject = mobject.current_dynamic_mobject
            if mobject.target_id is not None and self.source_graph.contains(mobject.id):
                self.source_graph.find_dynamic_mobject(mobject.id).target_id = mobject.target_id

        self.recover_points = {}
        self.recover_submobjects = {}
        for id in self.mobject_union:
            self.recover_points[id] = self.mobject_union[id].current_dynamic_mobject.points.copy()
            self.recover_submobjects[id] = self.mobject_union[id].current_dynamic_mobject.submobjects.copy()

        for id, mobject in self.target_mobjects.items():
            if id not in self.source_mobjects:
                mobject.current_dynamic_mobject.points = VMobject().points
                mobject.current_dynamic_mobject.submobjects = []
            else:
                mobject.current_dynamic_mobject.points = self.target_graph.find_dynamic_mobject(id).points
                mobject.current_dynamic_mobject.submobjects = self.target_graph.find_dynamic_mobject(id).direct_submobject_tree().copy().submobjects

        for id, mobject in self.source_mobjects.items():
            mobject.current_dynamic_mobject.points = self.source_graph.find_dynamic_mobject(id).points
            mobject.current_dynamic_mobject.submobjects = self.source_graph.find_dynamic_mobject(id).direct_submobject_tree().copy().submobjects
            self.scene.scene_add(mobject)

    def end_transforms(self):
        
        if self.is_transforming == False:
            return
        
        self.is_transforming = False

        for mobject in self.graph.root_mobjects:
            self.scene.scene_add(mobject.current_dynamic_mobject)
            #self.scene.scene_remove(mobject.current_dynamic_mobject)
            #self.scene.scene_add(mobject.current_dynamic_mobject)

        for id, mobject in self.mobject_union.items():
            mobject.current_dynamic_mobject.points = self.recover_points[mobject.id]
            mobject.current_dynamic_mobject.submobjects = self.recover_submobjects[mobject.id]
        
        self.create_progress_point()
        
    def create_progress_point(self):
        self.source_graph = self.graph.copy()
        self.target_graph = None
        self.source_mobjects = { mobject.id: mobject for mobject in self.graph.mobjects }
        self.active_mobjects = [ *self.graph.mobjects ]


class SceneProgressManager():

    def __init__(
        self,
        scene: Scene
    ):
        self.scene = scene
        self.progress_managers: Dict[Graph, GraphProgressManager] = {}

    def get_progress_manager(self, graph: DynamicMobjectGraph):

        if graph in self.progress_managers:
            return self.progress_managers[graph]
        
        raise Exception()
    
    def get_progress_manager_from_mobject(self, mobject: DynamicMobject):
        
        manager_from_graph = None
        if mobject.graph in self.progress_managers:
            manager_from_graph = self.progress_managers[mobject.graph]
        
        manager_from_active_mobject = None
        for progress_manager in self.progress_managers.values():
            if mobject.identity in progress_manager.active_mobjects:
                if manager_from_active_mobject is not None:
                    if manager_from_graph is None:
                        raise Exception()
                    
                manager_from_active_mobject = progress_manager
                
        if manager_from_graph is None and manager_from_active_mobject is None:
            raise Exception()
        
        if manager_from_graph is not None:
            return manager_from_graph
        
        if manager_from_active_mobject is not None:
            return manager_from_active_mobject

        raise Exception()

    def add_graph(self, graph: DynamicMobjectGraph):
        if not graph in self.progress_managers:
            self.progress_managers[graph] = GraphProgressManager(graph, self.scene)

        self.progress_managers[graph].create_progress_point()
    
    def add(self, mobject: DynamicMobject):
        self.add_graph(mobject.graph)

    def wait(self):
        for progress_manager in self.progress_managers.values():
            if not progress_manager.is_transforming:
                progress_manager.create_progress_point()


def attach_progress_interceptors(scene: Scene) -> SceneProgressManager:

    if hasattr(scene, "ATTACH_PROGRESS_INTERCEPTORS_FLAG"):
        return scene
    
    scene.ATTACH_PROGRESS_INTERCEPTORS_FLAG = True

    scene_progress_manager = SceneProgressManager(scene)
    AbstractDynamicTransform._scene_progress_manager = scene_progress_manager

    scene_add = scene.add
    scene_wait = scene.wait
    scene_remove = scene.remove

    def _add(*mobjects: Mobject) -> Scene:
        
        for mobject in mobjects:
            for m in extract_direct_dynamic_mobjects(mobject):
                if isinstance(m, DynamicMobject):
                    scene_progress_manager.add(m)

        return scene_add(*mobjects)
    
    def _wait(*args, **kwargs) -> None:
        scene_progress_manager.wait()
        scene_wait(*args, **kwargs)

    scene.add = _add
    scene.wait = _wait

    scene.scene_add = scene_add
    scene.scene_wait = scene_wait
    scene.scene_remove = scene_remove

    return scene_progress_manager


class AbstractDynamicTransform(Animation):

    _scene_progress_manager: SceneProgressManager | None = None
    _scene: Scene | None = None

    @staticmethod
    def scene_add(scene: Scene, mobject: Mobject):
        _scene_progress_manager = AbstractDynamicTransform._scene_progress_manager
        if _scene_progress_manager is not None:
            scene.scene_add(mobject)
        else:
            scene.add(mobject)

    @staticmethod
    def scene_remove(scene: Scene, mobject: Mobject):
        _scene_progress_manager = AbstractDynamicTransform._scene_progress_manager
        if _scene_progress_manager is not None:
            scene.scene_remove(mobject)
        else:
            scene.remove(mobject)

    @staticmethod
    def extract_subgraph(mobject: DynamicMobject | List[DynamicMobject] | DynamicMobjectSubgraph) -> DynamicMobjectSubgraph:

        if isinstance(mobject, DynamicMobjectSubgraph):
            return mobject

        if isinstance(mobject, DynamicMobject):
            return DynamicMobjectSubgraph.from_dynamic_mobject(mobject)
        
        mobjects: List[DynamicMobject] = []
        for item in mobject:
            mobjects.extend(AbstractDynamicTransform.extract_subgraph(item).dynamic_mobjects)

        return DynamicMobjectSubgraph(*mobjects)
    
    @classmethod
    def progress(
        cls,
        mobject: DynamicMobject | DynamicMobjectSubgraph | List[DynamicMobject | DynamicMobjectSubgraph | Any],
        **kwargs
    ):
        _scene_progress_manager = cls._scene_progress_manager
        
        if _scene_progress_manager is None:
            raise Exception("Missing attach_progress_interceptors(self) in the body of `def construct(self)`")
        

        mobjects: Set[DynamicMobject] = set()
        def extract_mobjects(mobject):

            if isinstance(mobject, DynamicMobjectSubgraph):
                for connected_mobject in mobject.dynamic_mobjects:
                    mobjects.add(connected_mobject)
                return

            if isinstance(mobject, DynamicMobject):
                for connected_mobject in DynamicMobjectSubgraph.from_dynamic_mobject(mobject).dynamic_mobjects:
                    mobjects.add(connected_mobject)
                return
            
            if isinstance(mobject, list):
                for item in mobject:
                    extract_mobjects(item)
                return
            
            raise Exception()
        
        """
        progress_manager = _scene_progress_manager.get_progress_manager(mobject.graph) """
        extract_mobjects(mobject)

        mobjects = list(mobjects)
        progress_manager, *other_progress_managers = [ _scene_progress_manager.get_progress_manager_from_mobject(mobject) for mobject in mobjects ]

        for other_progress_manager in other_progress_managers:
            if other_progress_manager is not  progress_manager:
                raise Exception()

        progress_manager.begin_transforms()

        source_graph, target_graph = progress_manager.source_graph, progress_manager.target_graph
        mobject_union = progress_manager.mobject_union
        
        progress_mobjects: Set[MobjectIdentity] = set()
        transform_descriptor = GraphTransformDescriptor(source_graph, target_graph)

        def extract_subgraph(mobject):
            return AbstractDynamicTransform.extract_subgraph(mobject)

        progress_subgraph = extract_subgraph(mobject)

        for mobject1 in progress_subgraph.mobjects:
            for id2, mobject2 in mobject_union.items():
                if transform_descriptor.is_continuous_ancestor(mobject1.id, id2):
                    progress_mobjects.add(mobject2)

        animation = cls(
            source_graph,
            target_graph,
            set(mobject.id for mobject in progress_mobjects),
            **kwargs
        )
        
        submobject_saves = {}
        points_saves = {}

        def find_mobject_from_mobject_union(id):
            for mobject in mobject_union.values():
                if mobject.id == id:
                    return mobject
                
            raise Exception()

        def begin_scene(scene: Scene):
            
            for mobject in progress_mobjects:
                mobject = mobject.current_dynamic_mobject
                submobject_saves[mobject] = mobject.submobjects
                points_saves[mobject] = mobject.points
                
                continuous_child_ids = animation.config.transform_descriptor.child_union_ids(mobject.id)

                child_intersection = []
                for id in continuous_child_ids:
                    if transform_descriptor.is_continuous_ancestor(mobject.id, id):
                        child_intersection.append(find_mobject_from_mobject_union(id))

                mobject.submobjects = [ animation.config.transform_containers[mobject.id], *child_intersection ]

        def clean_scene(scene: Scene):

            for mobject in progress_mobjects:
                mobject = mobject.current_dynamic_mobject
                mobject.submobjects = []

                if transform_descriptor.is_scene_remover(mobject.id):
                    scene.scene_add(mobject)
                    scene.scene_remove(mobject)

                if transform_descriptor.is_scene_introducer(mobject.id):
                    scene.scene_add(mobject)

                mobject.points = progress_manager.recover_points[mobject.id]
                mobject.submobjects = progress_manager.recover_submobjects[mobject.id]

        animation.setup_functions = [ begin_scene ]
        animation.clean_functions = [ clean_scene ]


        function_id = uuid.uuid4()

        def create_progress_point(graph: DynamicMobjectGraph):
            _scene_progress_manager = cls._scene_progress_manager

            for mobject in graph.dynamic_mobjects:
                mobject.source_id = None
                mobject.target_id = None

            if _scene_progress_manager is not None:
                _scene_progress_manager.add_graph(graph)

            graph.unsubscribe(function_id)

        progress_manager.graph.subscribe(lambda graph: create_progress_point(graph), function_id)


        return animation

    
    @classmethod
    def replacement_transform(
        cls,
        source_mobject: DynamicMobject,
        target_mobject: DynamicMobject,
        **kwargs
    ):
        source_graph = source_mobject.graph
        target_graph = target_mobject.graph

        subgraph1 = DynamicMobjectSubgraph.from_dynamic_mobject(source_mobject)
        subgraph2 = DynamicMobjectSubgraph.from_dynamic_mobject(target_mobject)
        union = { mobject.id for mobject in subgraph1.dynamic_mobjects }.union({ mobject.id for mobject in subgraph2.dynamic_mobjects })

        animation = cls(
            source_graph,
            target_graph, 
            union, 
            **kwargs
        )

        submobject_saves = {}
        points_saves = {}

        def begin_scene(scene: Scene):
            AbstractDynamicTransform._scene = scene

            cls.scene_add(scene, source_mobject)
            cls.scene_remove(scene, source_mobject)
    
            for mobject in subgraph1.dynamic_mobjects:
                
                submobject_saves[mobject] = mobject.submobjects
                points_saves[mobject] = mobject.points
                
                continuous_child_ids = animation.config.transform_descriptor.child_union_ids(mobject.id)
                mobject.submobjects = [ animation.config.transform_containers[mobject.id], *[ animation.config.transform_containers[id] for id in continuous_child_ids ]]

            for mobject in subgraph2.dynamic_mobjects:
                
                submobject_saves[mobject] = mobject.submobjects
                points_saves[mobject] = mobject.points
                
                continuous_child_ids = animation.config.transform_descriptor.child_union_ids(mobject.id)
                mobject.submobjects = [ animation.config.transform_containers[mobject.id], *[ animation.config.transform_containers[id] for id in continuous_child_ids ]]

        def clean_scene(scene: Scene):
            cls.scene_add(scene, target_mobject)

            for mobject in subgraph1.dynamic_mobjects:
                mobject.submobjects = submobject_saves[mobject]
                mobject.points = points_saves[mobject]

            for mobject in subgraph2.dynamic_mobjects:
                mobject.submobjects = submobject_saves[mobject]
                mobject.points = points_saves[mobject]

        animation.setup_functions = [ begin_scene ]
        animation.clean_functions = [ clean_scene ]


        function_id = uuid.uuid4()

        def create_progress_point(graph: DynamicMobjectGraph):
            _scene_progress_manager = cls._scene_progress_manager

            for mobject in graph.dynamic_mobjects:
                mobject.source_id = None
                mobject.target_id = None

            if _scene_progress_manager is not None:
                _scene_progress_manager.add_graph(graph)

            for root_mobject in graph.root_mobjects:
                cls.scene_add(AbstractDynamicTransform._scene, root_mobject)

            graph.unsubscribe(function_id)

        target_graph.subscribe(lambda graph: create_progress_point(graph), function_id)


        return animation
    
    
    @classmethod
    def from_copy(
        cls,
        source_mobject,
        target_mobject,
        **kwargs
    ):
        if not isinstance(source_mobject, DynamicMobjectGraph):
            source_graph = cls.extract_subgraph(source_mobject).graph
        else:
            source_graph = source_mobject

        target_subgraph = cls.extract_subgraph(target_mobject)
        target_graph = target_subgraph.graph

        extra_ids = set()
        for mobject in source_graph.mobjects:
            if target_subgraph.contains(mobject.current_dynamic_mobject.target_id):
                extra_ids.add(mobject.id)
            if target_subgraph.contains(mobject.current_dynamic_mobject.source_id):
                extra_ids.add(mobject.id)

        animation = cls(
            source_graph,
            target_graph,
            set(mobject.id for mobject in target_subgraph.mobjects).union(extra_ids),
            **kwargs
        )

        recover_submobjects = {}
        recover_points = {}

        def begin_scene(scene: Scene):
            AbstractDynamicTransform._scene = scene

            for mobject in target_subgraph.dynamic_mobjects:
                
                recover_submobjects[mobject] = mobject.submobjects
                recover_points[mobject] = mobject.points

                id_union = animation.config.transform_descriptor.child_union_ids(mobject.id)
                id_intersection = []

                for id in id_union:
                    if animation.config.transform_descriptor.is_continuous_ancestor(mobject.id, id):
                        id_intersection.append(id)

                mobject.submobjects = [ animation.config.transform_containers[mobject.id], *[ target_graph.find_dynamic_mobject(id) for id in id_intersection ]]

        def clean_scene(scene: Scene):
            
            for mobject in target_subgraph.dynamic_mobjects:
                
                mobject.submobjects = []
                cls.scene_add(scene, mobject)

                mobject.submobjects = recover_submobjects[mobject]
                mobject.points = recover_points[mobject]

        animation.setup_functions = [ begin_scene ]
        animation.clean_functions = [ clean_scene ]


        function_id = uuid.uuid4()

        def create_progress_point(graph: DynamicMobjectGraph):
            _scene_progress_manager = cls._scene_progress_manager

            for mobject in graph.dynamic_mobjects:
                mobject.source_id = None
                mobject.target_id = None

            if _scene_progress_manager is not None:
                _scene_progress_manager.add_graph(graph)

            for root_mobject in graph.root_mobjects:
                cls.scene_add(AbstractDynamicTransform._scene, root_mobject)

            graph.unsubscribe(function_id)

        target_graph.subscribe(lambda graph: create_progress_point(graph), function_id)
            

        return animation
    
    def __init__(
        self,
        source_graph: DynamicMobjectGraph, 
        target_graph: DynamicMobjectGraph,
        ids: Set[UUID],
        run_time: float | None = None, 
        **kwargs
    ):
        self.provided_run_time = run_time
        if run_time is None:
            run_time = 1

        self.config = DynamicTransformConfiguration(source_graph, target_graph, ids)
        self.container_group = VGroup(*self.config.transform_containers.values())

        self.setup_functions: List[Callable[[Scene], None]] = []
        self.clean_functions: List[Callable[[Scene], None]] = []

        self.apply()
        self.super_mobject = VMobject()
        super().__init__(mobject=self.super_mobject, run_time=run_time, **kwargs)

    @abstractmethod
    def apply(self):
        pass

    def intercept(
        self,
        mobject: DynamicMobject | DynamicMobjectSubgraph | UUID | List[Any]
    ) -> ItinerarySelectionInterceptor:

        ids: Set[UUID] = set()

        def recursive_extract(mobject):

            if isinstance(mobject, DynamicMobject):
                connected_ids = set(m.id for m in DynamicMobjectSubgraph.from_dynamic_mobject(mobject).mobjects)
                for connected_id in connected_ids:
                    ids.add(connected_id)
                return

            if isinstance(mobject, DynamicMobjectSubgraph):
                connected_ids = set(m.id for m in mobject.mobjects)
                for connected_id in connected_ids:
                    ids.add(connected_id)
                return

            if isinstance(mobject, UUID):
                ids.add(mobject)
                return

            if isinstance(mobject, list):
                for m in mobject:
                    recursive_extract(m)
                return

            raise Exception(
                f"Recieved {mobject} as input for animation intercept."
                "Only a DynamicMobject, DynamicMobjectSubgraph, UUID, or a list of those types, can be used to define ids to intercept." 
            )

        recursive_extract(mobject)
        return self.config.intercept(list(ids))

    def create_track(
        self,
        parent_track: AnimationTrack | None = None,
        start_time: float = 0,
        run_time: float = 1,
        name: str | None = None
    ) -> AnimationTrack:
        return self.config.create_track(parent_track=parent_track, start_time=start_time, run_time=run_time, name=name)

    def _setup_scene(self, scene):
        self.animation = self.config.build()
        
        if self.provided_run_time is not None:
            self.run_time = self.provided_run_time
        else:
            self.run_time = self.animation.run_time

        scene.add(self.animation.mobject)

        for function in self.setup_functions:
            function(scene)

    def begin(self):
        self.animation.begin()
        
    def finish(self):
        self.animation.finish()

    def clean_up_from_scene(self, scene: Scene) -> None:
        self.animation.clean_up_from_scene(scene)

        scene.remove(self.super_mobject)
        self.scene_add(scene, self.animation.mobject)
        self.scene_remove(scene, self.animation.mobject)

        for function in self.clean_functions:
            function(scene)
    
    def update_mobjects(self, dt: float) -> None:
        self.animation.update_mobjects(dt)

    def init_run_time(self, run_time) -> float:
        pass

    def build_animations_with_timings(self) -> None:
        pass

    def interpolate(self, alpha: float) -> None:
        self.animation.interpolate(alpha)

    
class GraphTransformDescriptor():

    def __init__(
        self, 
        source_graph: DynamicMobjectGraph,
        target_graph: DynamicMobjectGraph
    ):
        self.source_graph = source_graph
        self.target_graph = target_graph

    def find_source_dynamic_mobject(self, id: UUID) -> DynamicMobject | None:

        if self.source_graph.contains(id):
            return self.source_graph.find_dynamic_mobject(id)
        else:
            source_id = self.target_graph.get_dynamic_mobject(id).source_id
            if source_id is not None and self.source_graph.contains(source_id):
                return self.source_graph.find_dynamic_mobject(source_id)
            else:
                for mobject in self.source_graph.dynamic_mobjects:
                    if mobject.source_id == id:
                        return mobject
                
                if self.target_graph.contains(source_id):
                    for mobject in self.source_graph.dynamic_mobjects:
                        if mobject.source_id == source_id:
                            return mobject
                
                return None
    
    def find_target_dynamic_mobject(self, id: UUID) -> DynamicMobject | None:
        
        if self.target_graph.contains(id):
            return self.target_graph.find_dynamic_mobject(id)
        else:
            target_id = self.source_graph.get_dynamic_mobject(id).target_id
            if target_id is not None and self.target_graph.contains(target_id):
                return self.target_graph.find_dynamic_mobject(target_id)
            else:
                source_id = self.source_graph.find_dynamic_mobject(id).source_id
                for mobject in self.target_graph.dynamic_mobjects:
                    if mobject.id == source_id:
                        return mobject

                return None
    
    def has_source(self, id: UUID):
        return True if self.find_source_dynamic_mobject(id) is not None else False
    
    def has_target(self, id: UUID):
        return True if self.find_target_dynamic_mobject(id) is not None else False
    
    def is_remover(self, id: UUID):
        return self.has_source(id) and not self.has_target(id)
    
    def is_introducer(self, id: UUID):
        return self.has_target(id) and not self.has_source(id)
        
    def is_transformer(self, id: UUID):
        return not self.is_remover(id) and not self.is_introducer(id)
    
    def is_scene_remover(self, id: UUID):
        return self.source_graph.contains(id) and not self.target_graph.contains(id)
    
    def is_scene_introducer(self, id: UUID):
        return self.target_graph.contains(id) and not self.source_graph.contains(id)

    def is_continuous_ancestor(self, ancestor_id: UUID, child_id: UUID):
        
        a1 = self.has_source(ancestor_id) and self.has_target(ancestor_id)
        a2 = self.has_source(ancestor_id) and not a1
        a3 = self.has_target(ancestor_id) and not a1

        b1 = self.has_source(child_id) and self.has_target(child_id)
        b2 = self.has_source(child_id) and not b1
        b3 = self.has_target(child_id) and not b1
        
        if a2 and b2:
            return self.is_source_ancestor(ancestor_id, child_id)
        
        if a1 and b2:
            return self.is_source_ancestor(ancestor_id, child_id)
        
        if b3 and a3:
            return self.is_target_ancestor(ancestor_id, child_id)
        
        if b3 and a1:
            return self.is_target_ancestor(ancestor_id, child_id)
        
        if a1 and b1:
            return self.is_source_ancestor(ancestor_id, child_id) and self.is_target_ancestor(ancestor_id, child_id)
        
        return False

    def is_source_ancestor(self, ancestor_id: UUID, child_id: UUID):

        def recursive_has_ancestor_with_id(mobject: DynamicMobject):

            if mobject.id == ancestor_id:
                return True

            if mobject.parent:
                return recursive_has_ancestor_with_id(mobject.parent)
            else:
                return False

        source_child = self.find_source_dynamic_mobject(child_id)

        if source_child is None:
            return False
        else:
            return recursive_has_ancestor_with_id(source_child)
        
    def is_target_ancestor(self, ancestor_id: UUID, child_id: UUID):

        def recursive_has_ancestor_with_id(mobject: DynamicMobject):

            if mobject.id == ancestor_id:
                return True

            if mobject.parent:
                return recursive_has_ancestor_with_id(mobject.parent)
            else:
                return False

        target_child = self.find_target_dynamic_mobject(child_id)

        if target_child is None:
            return False
        else:
            return recursive_has_ancestor_with_id(target_child)
        
    def child_union_ids(self, id: UUID) -> List[UUID]:
        children: List[UUID] = set()

        if self.has_source(id):
            children = children.union(set([ mobject.id for mobject in self.find_source_dynamic_mobject(id).children ]))
        if self.has_target(id):
            children = children.union(set([ mobject.id for mobject in self.find_target_dynamic_mobject(id).children ]))

        return children


class ExtractTransform(Animation):

    def __init__(
        self,
        mobject: Mobject,
        animation: Animation
    ):
        self.super_mobject = VMobject()
        self.mobject = mobject
        self.animation = animation
        super().__init__(mobject=self.super_mobject)

    def begin(self):
        self.animation.begin()

    def finish(self):
        self.animation.finish()

    def interpolate(self, alpha: float) -> None:
        self.animation.interpolate(alpha)
        self.mobject.become(self.animation.mobject)

    def get_run_time(self) -> float:
        return self.animation.get_run_time()
        
    def clean_up_from_scene(self, scene: Scene) -> None:
        scene.remove(self.super_mobject)


class DynamicMobjectTransformItinerary():

    def __init__(
        self,
        config: DynamicTransformConfiguration,
        id: UUID
    ):
        self.config = config
        self.id = id

        self.source_mobject: VMobject = VMobject()
        self.target_mobject: VMobject = VMobject()
        self.animation_generator: Callable[[Mobject, Mobject], Animation] = lambda source, target: Transform(source, target)
        self.track = AnimationTrack(config, parent_track=self.config.root_track, run_time=1, is_leaf_track=True, name=self.id, id=self.id)

        if self.config.transform_descriptor.has_source(self.id):
            self.source_mobject = self.config.transform_descriptor.find_source_dynamic_mobject(self.id).direct_submobject_tree().copy()

        if self.config.transform_descriptor.has_target(self.id):
            self.target_mobject = self.config.transform_descriptor.find_target_dynamic_mobject(self.id).direct_submobject_tree().copy()
            
    def set_animation_generator(self, animation_generator):
        self.animation_generator = animation_generator

    def set_track(self, track: AnimationTrack):
        self.track.set_parent(track)

    def is_initialized(self):

        if self.source_mobject is not None and self.target_mobject is not None and self.animation_generator is not None and self.track is not None:
            return True
        
        return False

    def build(self):

        if not self.is_initialized():
            raise Exception()
        
        animation = self.animation_generator(self.source_mobject.copy(), self.target_mobject.copy())
        animation.run_time = self.track.run_time
        container_animation = ExtractTransform(self.config.transform_containers[self.id], animation)
        return container_animation


class ItinerarySelectionInterceptor():
    
    def __init__(
        self,
        config: DynamicTransformConfiguration,
        ids: List[uuid.UUID] 
    ):
        self.config = config

        for id in ids.copy():
            if id not in config.itineraries:
                ids.remove(id)    

        self.itineraries = { id: config.itineraries[id] for id in ids }

    def set_animation(self, animation_generator) -> ItinerarySelectionInterceptor:
        for itinerary in self.itineraries.values():
            itinerary.set_animation_generator(animation_generator)
        return self

    def set_track(self, track: AnimationTrack) -> ItinerarySelectionInterceptor:
        for itinerary in self.itineraries.values():
            itinerary.set_track(track)
        return self

    def set_source(self, mobject: Mobject | DynamicMobject) -> ItinerarySelectionInterceptor:

        if isinstance(mobject, DynamicMobject):
            for descendant in mobject.get_dynamic_family():
                if descendant.id in self.itineraries:
                    self.itineraries[descendant.id].source_mobject = descendant.direct_submobject_tree().copy()
        else:
            for itinerary in self.itineraries.values():
                itinerary.source_mobject = mobject.copy()

        return self

    def set_target(self, mobject: Mobject | DynamicMobject) -> ItinerarySelectionInterceptor:

        if isinstance(mobject, DynamicMobject):
            for mobject in mobject.get_dynamic_family():
                if mobject.id in self.itineraries:
                    self.itineraries[mobject.id].target_mobject = mobject.direct_submobject_tree().copy()
        else:
            for itinerary in self.itineraries.values():
                itinerary.target_mobject = mobject.copy()
                
        return self
    

class AnimationTrack():
    
    def set_run_time(self):
        if self.is_leaf_track and self.parent_track is not None:
            self.run_time = self.parent_track.run_time

    def __init__(
        self, 
        config: DynamicTransformConfiguration,
        start_time: float = 0.0, 
        run_time: float | None = None, 
        require_points: bool = True,
        is_root_track: bool = False,
        is_leaf_track: bool = False,
        parent_track: AnimationTrack | None = None,
        name: str | None = None,
        id: UUID | None = None, 
    ):
        #self.start_time = start_time
        self.run_time = run_time
        self.require_points = require_points

        self.config = config
        self.is_root_track = is_root_track
        self.is_leaf_track = is_leaf_track
        self.id = id

        if is_leaf_track and id is None:
            raise Exception()

        self.children: List[AnimationTrack] = []
        self.parent_track: AnimationTrack | None = None
        
        if parent_track:
            self.set_parent(parent_track)

        self.name = name
        self.start_time = start_time
       
    def set_parent(
        self, 
        parent_track: AnimationTrack | None
    ):
        if parent_track is not None:
            if parent_track.is_leaf_track:
                raise Exception()
        
            if self.config is not parent_track.config:
                raise Exception()

        if self.is_root_track:
            raise Exception()
        
        if self.parent_track:
            self.parent_track.children.remove(self)
        
        self.start_time = 0
        self.parent_track = parent_track

        if self.parent_track is not None:
            self.parent_track.children.append(self)

        self.set_run_time()

    def set_lag_ratio(self, lag_ratio: float):
        
        curr_time: float = 0
        for track in self.children:
            start_time: float = curr_time
            track.start_time = start_time

            end_time: float = start_time + track.run_time
            curr_time = (1 - lag_ratio) * start_time + lag_ratio * end_time

        self.run_time = curr_time

    def get_run_time(self):
        self.set_run_time()

        if not self.children and not self.is_leaf_track:
            return 0
        
        if self.is_leaf_track:
            return self.parent_track.run_time

        if self.run_time:
            return self.run_time

        _max = 0
        for track in self.children:
            _max = max(_max, track.start_time + track.get_run_time())

        return _max
    
    def get_itinerary(self):
        self.set_run_time()
        return self.config.itineraries[self.id]
    
    def has_mobject_with_points(self):
        self.set_run_time()
        if self.is_leaf_track:
            itinerary = self.get_itinerary()
            source_family = itinerary.source_mobject.get_family()
            target_family = itinerary.target_mobject.get_family()

            for mobject in source_family:
                if mobject.has_points():
                    return True
                
            for mobject in target_family:
                if mobject.has_points():
                    return True
                
            return False
        else:
            for track in self.children:
                if track.has_mobject_with_points():
                    return True
            
            return False


class ExactTimingsAnimationGroup(PreviousAnimationGroup):


    def __init__(
        self,
        *animations: Tuple[Animation, float],
        group = None,
        run_time: float | None = None,
        rate_func: Callable[[float], float] = linear,
        lag_ratio: float = 0,
        **kwargs,
    ):
        self.start_times = [ item[1] for item in animations ]     
        super().__init__(*[ item[0] for item in animations ])
        

    def build_animations_with_timings(self) -> None:
        self.anims_with_timings = []

        for start_time, anim in zip(self.start_times, self.animations):
            end_time: float = start_time + anim.get_run_time()
            self.anims_with_timings.append((anim, start_time, end_time))


class DynamicTransformConfiguration():

    def __init__(
        self,
        source_graph: DynamicMobjectGraph, 
        target_graph: DynamicMobjectGraph,
        ids: Set[UUID]
    ):
        self.source_graph = source_graph
        self.target_graph = target_graph

        for id in ids:
            if not (source_graph.contains(id) or target_graph.contains(id)):
                raise Exception()

        self.root_track = AnimationTrack(config=self, run_time=1, is_root_track=True, name="root_track")
        self.transform_descriptor = GraphTransformDescriptor(source_graph, target_graph)
        self.prevent_ids: Set[UUID] = set()
        
        self.ids = ids.copy()
        for id in ids:
            if self.transform_descriptor.is_remover(id):
                for mobject in self.target_graph.mobjects:
                    if mobject.current_dynamic_mobject.source_id is not None and mobject.current_dynamic_mobject.source_id == id:
                        if id in self.ids:
                            self.prevent_ids.add(id)
        

        self.transform_containers = { id: VMobject() for id in self.ids }
        self.itineraries = { id: DynamicMobjectTransformItinerary(self, id) for id in self.ids }

    def initialized(self):

        for itinerary in self.itineraries.values():
            if not itinerary.initialized():
                return False
            
        return True
    
    def create_track(
        self,
        parent_track: AnimationTrack | None = None,
        start_time: float = 0,
        run_time: float = 1,
        name: str | None = None
    ) -> AnimationTrack:
        
        if parent_track is None:
            parent_track = self.root_track

        return AnimationTrack(config=self, parent_track=parent_track, start_time=start_time, run_time=run_time, name=name)

    def build(self):
        
        def recursive_build(track: AnimationTrack) -> ExactTimingsAnimationGroup:
            

            if track.is_leaf_track:
                subtracks = [ ( track.get_itinerary().build() , 0) ]
            else:
                subtracks = [ ( recursive_build(track), track.start_time ) for track in track.children ]

            animation = ExactTimingsAnimationGroup(*subtracks, run_time=track.get_run_time())
            return animation
        
        animation = recursive_build(self.root_track)
        return animation

    def intercept(
        self, 
        ids: List[UUID]
    ):
        return ItinerarySelectionInterceptor(self, ids)


class TransformInStages(AbstractDynamicTransform):

    def __init__(
        self, 
        source_graph: DynamicMobjectGraph,
        target_graph: DynamicMobjectGraph,
        ids: Set[UUID],
        lag_ratio: float = 1,
        track_run_time: float = 1,
        **kwargs
    ):
        self._lag_ratio = lag_ratio
        self.track_run_time = track_run_time
        super().__init__(source_graph, target_graph, ids, **kwargs)
    
    def apply(self):

        self.default_track = self.config.create_track(name="default")
        self.remover_track = self.config.create_track(parent_track=self.default_track, name="default-remover", run_time=self.track_run_time)
        self.transformer_track = self.config.create_track(parent_track=self.default_track, name="default-transformer", run_time=self.track_run_time)
        self.introducer_track = self.config.create_track(parent_track=self.default_track, name="default-introducer", run_time=self.track_run_time)

        for itinerary in self.config.itineraries.values():
            itinerary.set_animation_generator(lambda source, target: Transform(source, target))

            if self.config.transform_descriptor.is_remover(itinerary.id):
                itinerary.target_mobject = self.config.source_graph.find_dynamic_mobject(itinerary.id).direct_submobject_tree().copy().fade(1)
                itinerary.set_track(self.remover_track)

            if self.config.transform_descriptor.is_introducer(itinerary.id):
                itinerary.source_mobject = self.config.target_graph.find_dynamic_mobject(itinerary.id).direct_submobject_tree().copy().fade(1)
                itinerary.set_track(self.introducer_track)

            if self.config.transform_descriptor.is_transformer(itinerary.id):
                itinerary.set_track(self.transformer_track)

            if itinerary.id in self.config.prevent_ids:
                itinerary.track.set_parent(None)
                itinerary.source_mobject = VMobject()
                itinerary.target_mobject = VMobject()

        if not self.remover_track.has_mobject_with_points():
            self.remover_track.set_parent(None)

        if not self.transformer_track.has_mobject_with_points():
            self.transformer_track.set_parent(None)
        
        if not self.introducer_track.has_mobject_with_points():
            self.introducer_track.set_parent(None)
        
        self.default_track.set_lag_ratio(self._lag_ratio)


import manim
from .animation import attach_progress_interceptors


class Scene(manim.Scene):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        attach_progress_interceptors(self)
        